@@ -42,7 +42,7 @@ import org.apache.lucene.util.FieldCacheSanityChecker;
  * <p>Created: May 19, 2004 4:40:36 PM
  *
  * @since   lucene 1.4
- * @version $Id: FieldCacheImpl.java 807572 2009-08-25 11:44:45Z mikemccand $
+ * @version $Id: FieldCacheImpl.java 950451 2010-06-02 09:33:57Z mikemccand $
  */
  // TODO: change interface to FieldCache in 3.0 when removed
 class FieldCacheImpl implements ExtendedFieldCache {
@@ -68,6 +68,14 @@ class FieldCacheImpl implements ExtendedFieldCache {
   public void purgeAllCaches() {
     init();
   }
+
+  public void purge(IndexReader r) {
+    Iterator it = caches.values().iterator();
+    while(it.hasNext()) {
+      Cache c = (Cache) it.next();
+      c.purge(r);
+    }
+  }
   
   public CacheEntry[] getCacheEntries() {
     List result = new ArrayList(17);
@@ -183,6 +191,14 @@ class FieldCacheImpl implements ExtendedFieldCache {
     protected abstract Object createValue(IndexReader reader, Entry key)
         throws IOException;
 
+    /** Remove this reader from the cache, if present. */
+    public void purge(IndexReader r) {
+      Object readerKey = r.getFieldCacheKey();
+      synchronized(readerCache) {
+        readerCache.remove(readerKey);
+      }
+    }
+
     public Object get(IndexReader reader, Entry key) throws IOException {
       Map innerCache;
       Object value;
@@ -702,10 +718,6 @@ class FieldCacheImpl implements ExtendedFieldCache {
           if (term==null || term.field() != field) break;
 
           // store term text
-          // we expect that there is at most one term per document
-          if (t >= mterms.length) throw new RuntimeException ("there are more terms than " +
-                  "documents in field \"" + field + "\", but it's impossible to sort on " +
-                  "tokenized fields");
           mterms[t] = term.text();
 
           termDocs.seek (termEnum);
