@@ -33,6 +33,7 @@ import org.apache.lucene.document.Field.TermVector;
 import org.apache.lucene.search.TermQuery;
 import org.apache.lucene.search.IndexSearcher;
 import org.apache.lucene.search.Query;
+import org.apache.lucene.search.TopDocs;
 import org.apache.lucene.store.Directory;
 import org.apache.lucene.store.MockRAMDirectory;
 import org.apache.lucene.store.AlreadyClosedException;
@@ -222,6 +223,7 @@ public class TestIndexWriterReader extends LuceneTestCase {
     Directory dir1 = new MockRAMDirectory();
     IndexWriter writer = new IndexWriter(dir1, new WhitespaceAnalyzer(),
         IndexWriter.MaxFieldLength.LIMITED);
+    writer.setReaderTermsIndexDivisor(2);
     writer.setInfoStream(infoStream);
     // create the index
     createIndexNoClose(!optimize, "index1", writer);
@@ -616,7 +618,7 @@ public class TestIndexWriterReader extends LuceneTestCase {
     writer.setMergeFactor(2);
     writer.setMaxBufferedDocs(2);
 
-    for (int i = 0; i < 10; i++) {
+    for (int i = 0; i < 100; i++) {
       writer.addDocument(createDocument(i, "test", 4));
     }
     ((ConcurrentMergeScheduler) writer.getMergeScheduler()).sync();
@@ -889,4 +891,26 @@ public class TestIndexWriterReader extends LuceneTestCase {
     w.close();
     dir.close();
   }
+  public void testSegmentWarmer() throws Exception {
+    Directory dir = new MockRAMDirectory();
+    IndexWriter w = new IndexWriter(dir, new WhitespaceAnalyzer(), IndexWriter.MaxFieldLength.UNLIMITED);
+    w.setMaxBufferedDocs(2);
+    w.getReader().close();
+    w.setMergedSegmentWarmer(new IndexWriter.IndexReaderWarmer() {
+        public void warm(IndexReader r) throws IOException {
+          final IndexSearcher s = new IndexSearcher(r);
+          final TopDocs hits = s.search(new TermQuery(new Term("foo", "bar")), 10);
+          assertEquals(20, hits.totalHits);
+        }
+      });
+    
+    Document doc = new Document();
+    doc.add(new Field("foo", "bar", Field.Store.YES, Field.Index.NOT_ANALYZED));
+    for(int i=0;i<20;i++) {
+      w.addDocument(doc);
+    }
+    w.waitForMerges();
+    w.close();
+    dir.close();
+  }
 }
